# Standard Library
import datetime
import random
from typing import Dict, List, Optional

# Third Party Code
from bitstring import BitStream
from dateutil.tz import tzlocal
import ephem

# Supercell Code
from supercell.weatherlink.exceptions import BadCRC
from supercell.weatherlink.utils import crc16, CRC16_TABLE, make_time

FORECAST_RULES = [
    "Mostly clear and cooler.",
    "Mostly clear with little temperature change.",
    "Mostly clear for 12 hours with little temperature change.",
    "Mostly clear for 12 to 24 hours and cooler.",
    "Mostly clear with little temperature change.",
    "Partly cloudy and cooler.",
    "Partly cloudy with little temperature change.",
    "Partly cloudy with little temperature change.",
    "Mostly clear and warmer.",
    "Partly cloudy with little temperature change.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and warmer. Precipitation possible within 24 to 48 hours.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds with little temperature change. Precipitation possible within 24 hours.",
    "Mostly clear with little temperature change.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds with little temperature change. Precipitation possible within 12 hours.",
    "Mostly clear with little temperature change.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and warmer. Precipitation possible within 24 hours.",
    "Mostly clear and warmer. Increasing winds.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and warmer. Precipitation possible within 12 hours. Increasing winds.",
    "Mostly clear and warmer. Increasing winds.",
    "Increasing clouds and warmer.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and warmer. Precipitation possible within 12 hours. Increasing winds.",
    "Mostly clear and warmer. Increasing winds.",
    "Increasing clouds and warmer.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and warmer. Precipitation possible within 12 hours. Increasing winds.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Mostly clear and warmer. Precipitation possible within 48 hours.",
    "Mostly clear and warmer.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds with little temperature change. Precipitation possible within 24 to 48 hours.",
    "Increasing clouds with little temperature change.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and warmer. Precipitation possible within 12 to 24 hours.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and warmer. Precipitation possible within 12 to 24 hours. Windy.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and warmer. Precipitation possible within 12 to 24 hours. Windy.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and warmer. Precipitation possible within 6 to 12 hours.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and warmer. Precipitation possible within 6 to 12 hours. Windy.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and warmer. Precipitation possible within 12 to 24 hours. Windy.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and warmer. Precipitation possible within 12 hours.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and warmer. Precipitation likley.",
    "Clearing and cooler. Precipitation ending within 6 hours.",
    "Partly cloudy with little temperature change.",
    "Clearing and cooler. Precipitation ending within 6 hours.",
    "Mostly clear with little temperature change.",
    "Clearing and cooler. Precipitation ending within 6 hours.",
    "Partly cloudy and cooler.",
    "Partly cloudy with little temperature change.",
    "Mostly clear and cooler.",
    "Clearing and cooler. Precipitation ending within 6 hours.",
    "Mostly clear with little temperature change.",
    "Clearing and cooler. Precipitation ending within 6 hours.",
    "Mostly clear and cooler.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds with little temperature change. Precipitation possible within 24 hours.",
    "Mostly cloudy and cooler. Precipitation continuing.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Mostly cloudy and cooler. Precipitation likely.",
    "Mostly cloudy with little temperature change. Precipitation continuing.",
    "Mostly cloudy with little temperature change. Precipitation likely.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and cooler. Precipitation possible and windy within 6 hours.",
    "Increasing clouds with little temperature change. Precipitation possible and windy within 6 hours.",
    "Mostly cloudy and cooler. Precipitation continuing. Increasing winds.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Mostly cloudy and cooler. Precipitation likely. Increasing winds.",
    "Mostly cloudy with little temperature change. Precipitation continuing. Increasing winds.",
    "Mostly cloudy with little temperature change. Precipitation likely. Increasing winds.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and cooler. Precipitation possible within 12 to 24 hours possible wind shift to the W NW or N.",
    "Increasing clouds with little temperature change. Precipitation possible within 12 to 24 hours possible wind "
    "shift to the W NW or N.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and cooler. Precipitation possible within 6 hours possible wind shift to the W NW or N.",
    "Increasing clouds with little temperature change. Precipitation possible within 6 hours possible wind shift to "
    "the W NW or N.",
    "Mostly cloudy and cooler. Precipitation ending within 12 hours possible wind shift to the W NW or N.",
    "Mostly cloudy and cooler. Possible wind shift to the W NW or N.",
    "Mostly cloudy with little temperature change. Precipitation ending within 12 hours possible wind shift to the W "
    "NW or N.",
    "Mostly cloudy with little temperature change. Possible wind shift to the W NW or N.",
    "Mostly cloudy and cooler. Precipitation ending within 12 hours possible wind shift to the W NW or N.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Mostly cloudy and cooler. Precipitation possible within 24 hours possible wind shift to the W NW or N.",
    "Mostly cloudy with little temperature change. Precipitation ending within 12 hours possible wind shift to the W "
    "NW or N.",
    "Mostly cloudy with little temperature change. Precipitation possible within 24 hours possible wind shift to the "
    "W NW or N.",
    "Clearing cooler and windy. Precipitation ending within 6 hours.",
    "Clearing cooler and windy.",
    "Mostly cloudy and cooler. Precipitation ending within 6 hours. Windy with possible wind shift to the W NW or N.",
    "Mostly cloudy and cooler. Windy with possible wind shift to the W NW or N.",
    "Clearing cooler and windy.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Mostly cloudy with little temperature change. Precipitation possible within 12 hours. Windy.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and cooler. Precipitation possible within 12 hours possibly heavy at times. Windy.",
    "Mostly cloudy and cooler. Precipitation ending within 6 hours. Windy.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Mostly cloudy and cooler. Precipitation possible within 12 hours. Windy.",
    "Mostly cloudy and cooler. Precipitation ending in 12 to 24 hours.",
    "Mostly cloudy and cooler.",
    "Mostly cloudy and cooler. Precipitation continuing possible heavy at times. Windy.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Mostly cloudy and cooler. Precipitation possible within 6 to 12 hours. Windy.",
    "Mostly cloudy with little temperature change. Precipitation continuing possibly heavy at times. Windy.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Mostly cloudy with little temperature change. Precipitation possible within 6 to 12 hours. Windy.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds with little temperature change. Precipitation possible within 12 hours possibly heavy at "
    "times. Windy.",
    "Mostly cloudy and cooler. Windy.",
    "Mostly cloudy and cooler. Precipitation continuing possibly heavy at times. Windy.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Mostly cloudy and cooler. Precipitation likely possibly heavy at times. Windy.",
    "Mostly cloudy with little temperature change. Precipitation continuing possibly heavy at times. Windy.",
    "Mostly cloudy with little temperature change. Precipitation likely possibly heavy at times. Windy.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and cooler. Precipitation possible within 6 hours. Windy.",
    "Increasing clouds with little temperature change. Precipitation possible within 6 hours. Windy",
    "Increasing clouds and cooler. Precipitation continuing. Windy with possible wind shift to the W NW or N.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Mostly cloudy and cooler. Precipitation likely. Windy with possible wind shift to the W NW or N.",
    "Mostly cloudy with little temperature change. Precipitation continuing. Windy with possible wind shift to the W "
    "NW or N.",
    "Mostly cloudy with little temperature change. Precipitation likely. Windy with possible wind shift to the W NW "
    "or N.",
    "Increasing clouds and cooler. Precipitation possible within 6 hours. Windy with possible wind shift to the W NW "
    "or N.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and cooler. Precipitation possible within 6 hours possible wind shift to the W NW or N.",
    "Increasing clouds with little temperature change. Precipitation possible within 6 hours. Windy with possible "
    "wind shift to the W NW or N.",
    "Increasing clouds with little temperature change. Precipitation possible within 6 hours possible wind shift to "
    "the W NW or N.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and cooler. Precipitation possible within 6 hours. Windy with possible wind shift to the W NW "
    "or N.",
    "Increasing clouds with little temperature change. Precipitation possible within 6 hours. Windy with possible "
    "wind shift to the W NW or N.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Increasing clouds and cooler. Precipitation possible within 12 to 24 hours. Windy with possible wind shift to "
    "the W NW or N.",
    "Increasing clouds with little temperature change. Precipitation possible within 12 to 24 hours. Windy with "
    "possible wind shift to the W NW or N.",
    "Mostly cloudy and cooler. Precipitation possibly heavy at times and ending within 12 hours. Windy with possible "
    "wind shift to the W NW or N.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Mostly cloudy and cooler. Precipitation possible within 6 to 12 hours possibly heavy at times. Windy with "
    "possible wind shift to the W NW or N.",
    "Mostly cloudy with little temperature change. Precipitation ending within 12 hours. Windy with possible wind "
    "shift to the W NW or N.",
    "Mostly cloudy with little temperature change. Precipitation possible within 6 to 12 hours possibly heavy at "
    "times. Windy with possible wind shift to the W NW or N.",
    "Mostly cloudy and cooler. Precipitation continuing.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Mostly cloudy and cooler. Precipitation likely. Windy with possible wind shift to the W NW or N.",
    "Mostly cloudy with little temperature change. Precipitation continuing.",
    "Mostly cloudy with little temperature change. Precipitation likely.",
    "Partly cloudy with little temperature change.",
    "Mostly clear with little temperature change.",
    "Mostly cloudy and cooler. Precipitation possible within 12 hours possibly heavy at times. Windy.",
    "FORECAST REQUIRES 3 HOURS OF RECENT DATA",
    "Mostly clear and cooler.",
    "Mostly clear and cooler.",
    "Mostly clear and cooler.",
    "Unknown forecast rule."]

BAR_TREND_CODE_FALLING_RAPIDLY = -60
BAR_TREND_TEXT_FALLING_RAPIDLY = "Falling Rapidly"

BAR_TREND_CODE_FALLING_SLOWLY = -20
BAR_TREND_TEXT_FALLING_SLOWLY = "Falling Slowly"

BAR_TREND_CODE_STABLE = 0
BAR_TREND_TEXT_STABLE = "Stable"

BAR_TREND_CODE_RISING_SLOWLY = 20
BAR_TREND_TEXT_RISING_SLOWLY = "Rising Slowly"

BAR_TREND_CODE_RISING_RAPIDLY = 60
BAR_TREND_TEXT_RISING_RAPIDLY = "Rising Rapidly"

BAR_TREND_TEXT = dict((
    (BAR_TREND_CODE_FALLING_RAPIDLY, BAR_TREND_TEXT_FALLING_RAPIDLY),
    (BAR_TREND_CODE_FALLING_SLOWLY, BAR_TREND_TEXT_FALLING_SLOWLY),
    (BAR_TREND_CODE_STABLE, BAR_TREND_TEXT_STABLE),
    (BAR_TREND_CODE_RISING_SLOWLY, BAR_TREND_TEXT_RISING_SLOWLY),
    (BAR_TREND_CODE_RISING_RAPIDLY, BAR_TREND_TEXT_RISING_RAPIDLY),
))

LOOP_RECORD_SIZE_BYTES = 99
LOOP_RECORD_SIZE_BITS = LOOP_RECORD_SIZE_BYTES * 8

REVISION_A = "A"
REVISION_B = "B"

LOOP_HEADER = b"LOOP"

LOOP2_PACKET_TYPE = "LOOP2"
LOOP_PACKET_TYPE = "LOOP"

LUNATION_LOOKUP = {
    0.05: "New Moon",
    0.15: "Crescent",
    0.25: "First Quarter",
    0.45: "Gibbous",
    0.55: "Full Moon",
    0.65: "Gibbous",
    0.85: "Last Quarter",
    0.95: "Crescent",
    1.00: "New Moon"
}


WIND_DIRECTION_LOOKUP = {
    10: "N",
    35: "NNE",
    55: "NE",
    80: "NEE",
    110: "E",
    125: "SEE",
    145: "SE",
    170: "SSE",
    190: "S",
    215: "SSW",
    235: "SW",
    260: "SWW",
    280: "W",
    305: "NWW",
    325: "NW",
    350: "NNW",
    360: "N"
}

FORECAST_ICONS_LOOKUP = {
    1: "Rain within 12 hrs",
    2: "Cloudy",
    4: "Cloud + Sunny",
    8: "Sunny",
    16: "Snow within 12hrs"
}


def _lunation_text(lunation: float) -> str:
    """Converts the lunation value to a string."""
    if lunation < 0 or lunation > 1.0:
        raise ValueError("Lunation must be between 0.0 - 1.0")

    if lunation < 0.5:
        direction = "Waxing"
    else:
        direction = "Waning"

    for lunation_value, lunation_text in LUNATION_LOOKUP.items():
        if lunation >= lunation_value:
            continue
        break

    return "{text} ({direction})".format(
        text=lunation_text,
        direction=direction
    )


def _wind_direction_text(wind_direction: int) -> str:
    """Converts the wind direction value to text."""
    if wind_direction < 0 or wind_direction > 360:
        raise ValueError("Wind direction (%s) must be between 0 - 360" % (wind_direction))
    for wind_direction_value, wind_direction_text in WIND_DIRECTION_LOOKUP.items():
        if wind_direction >= wind_direction_value:
            continue
        break
    return wind_direction_text


def _forecast_icons_text(forecast_icons: int) -> List[str]:
    if forecast_icons < 0 or forecast_icons > 31:
        raise ValueError("Forecast icons must be between 0 - 31.")
    return [
        forecast_icon_text
        for bit_position, forecast_icon_text
        in FORECAST_ICONS_LOOKUP.items()
        if ((forecast_icons & bit_position) == bit_position)
    ]


def get_phase_on_day(year: int, month: int, day: int) -> float:
    """Returns a floating-point number from 0-1. where 0=new, 0.5=full, 1=new"""
    date = ephem.Date(datetime.date(year, month, day))
    nnm = ephem.next_new_moon(date)
    pnm = ephem.previous_new_moon(date)
    lunation = (date - pnm) / (nnm - pnm)
    return lunation


class StationObservation(object):
    """A station observation"""
    bar_trend: int
    barometer: float
    inside_temperature: float
    inside_humidity: float
    outside_temperature: float
    outside_humidity: float
    wind_speed: int
    ten_min_avg_wind_speed: int
    wind_direction: int
    rain_rate: int
    console_battery_voltage: float
    forecast_icons: int
    forecast_rule_number: int
    sunrise: datetime.time
    sunset: datetime.time
    observation_made_at: datetime.datetime
    identitier: int

    def __init__(self,
                 bar_trend: int,
                 barometer: float,
                 inside_temperature: float,
                 inside_humidity: float,
                 outside_temperature: float,
                 outside_humidity: float,
                 wind_speed: int,
                 ten_min_avg_wind_speed: int,
                 wind_direction: int,
                 rain_rate: int,
                 console_battery_voltage: float,
                 forecast_icons: int,
                 forecast_rule_number: int,
                 sunrise: datetime.time,
                 sunset: datetime.time,
                 observation_made_at: Optional[datetime.datetime] = None,
                 identifier: Optional[int] = None
                 ) -> None:
        self.bar_trend = int(bar_trend)
        self.barometer = float(barometer)
        self.inside_temperature = float(inside_temperature)
        self.inside_humidity = float(inside_humidity)
        self.outside_temperature = float(outside_temperature)
        self.outside_humidity = float(outside_humidity)
        self.wind_speed = int(wind_speed)
        self.ten_min_avg_wind_speed = int(ten_min_avg_wind_speed)
        self.wind_direction = int(wind_direction)
        self.rain_rate = int(rain_rate)
        self.console_battery_voltage = float(console_battery_voltage)
        self.forecast_icons = int(forecast_icons)
        self.forecast_rule_number = int(forecast_rule_number)
        self.sunrise = sunrise
        self.sunset = sunset
        self.observation_made_at = observation_made_at or datetime.datetime.now(tzlocal())
        self.lunation = get_phase_on_day(
            self.observation_made_at.year,
            self.observation_made_at.month,
            self.observation_made_at.day)
        self.identifier = identifier or random.getrandbits(32)

    def lunation_text(self) -> str:
        return _lunation_text(self.lunation)

    def wind_direction_text(self) -> str:
        """Produces a string description of the wind direction."""
        return _wind_direction_text(self.wind_direction)

    def bar_trend_text(self) -> str:
        """Returns the string version of the bar trend."""
        return BAR_TREND_TEXT[self.bar_trend]

    def forecast_icons_text(self) -> List[str]:
        return _forecast_icons_text(self.forecast_icons)

    def forecast_text(self) -> str:
        """Returns the string version of the forecast rule."""
        return FORECAST_RULES[self.forecast_rule_number]

    def to_dict(self) -> Dict:
        """A dictionary representation of the observation."""
        return {
            "bar_trend": self.bar_trend,
            "bar_trend_text": self.bar_trend_text(),
            "barometer": self.barometer,
            "inside_temperature": self.inside_temperature,
            "inside_humidity": self.inside_humidity,
            "outside_temperature": self.outside_temperature,
            "outside_humidity": self.outside_humidity,
            "wind_speed": self.wind_speed,
            "ten_min_avg_wind_speed": self.ten_min_avg_wind_speed,
            "wind_direction": self.wind_direction,
            "wind_direction_text": self.wind_direction_text(),
            "rain_rate": self.rain_rate,
            "console_battery_voltage": self.console_battery_voltage,
            "forecast_icons": self.forecast_icons,
            "forecast_icons_text": self.forecast_icons_text(),
            "forecast_rule_number": self.forecast_rule_number,
            "forecast_text": self.forecast_text(),
            "sunrise": self.sunrise.isoformat(),
            "sunset": self.sunset.isoformat(),
            "lunation": self.lunation,
            "lunation_text": self.lunation_text(),
            "observation_made_at": self.observation_made_at,
            "identifier": self.identifier
        }

    @classmethod
    def validate_record(cls, record_bitstream: BitStream) -> None:
        """Validates a record."""
        if len(record_bitstream) != LOOP_RECORD_SIZE_BITS:
            raise ValueError("Records should be %d bytes in length." % LOOP_RECORD_SIZE_BYTES)
        if crc16(record_bitstream, 0, CRC16_TABLE) != 0:
            raise BadCRC()

    @classmethod
    def validate_packet_type(cls, record_bitstream: BitStream) -> None:
        """Validates the packet type."""
        packet_type = record_bitstream.read(8).int and "LOOP2" or "LOOP"
        if packet_type == LOOP2_PACKET_TYPE:
            raise ValueError("LOOP2 Packet not yet supported.")

    @classmethod
    def init_with_bytes(cls, record_bytes: bytes, identifier: Optional[int] = None,
                        observation_made_at: Optional[datetime.datetime] = None):
        """Creates a new Station Observation from record of bytes."""
        record_bitstream = BitStream(record_bytes)
        cls.validate_record(record_bitstream)
        # Set to position four
        record_bitstream.pos = 24
        # Awkwardly positioned bar trend
        bar_trend = record_bitstream.read(8).intle
        cls.validate_packet_type(record_bitstream)
        # Skip 2 bytes
        record_bitstream.read(16)

        barometer = record_bitstream.read(16).uintle / 1000.0
        inside_temperature = record_bitstream.read(16).intle / 10.0
        inside_humidity = record_bitstream.read(8).uintle
        outside_temperature = record_bitstream.read(16).intle / 10.0
        wind_speed = record_bitstream.read(8).uintle
        ten_min_avg_wind_speed = record_bitstream.read(8).uintle
        wind_direction = record_bitstream.read(16).uintle  # 0º = None, 90º = E, 180 = S, 270 = W, 360 = N

        # Skip "extra temperatures"
        record_bitstream.read(56)  # Each byte is a one extra temperature value in whole degrees F
        # with # an offset  of 90 degrees. 0 = -90, 100 = 10, 169 = 79
        # Skip soil temperatures
        record_bitstream.read(32)
        # Skip leaf temperatures
        record_bitstream.read(32)

        outside_humidity = record_bitstream.read(8).uintle

        # Skip extra humidities
        record_bitstream.read(56)

        rain_rate = record_bitstream.read(16).uintle

        # ultraviolet_index
        record_bitstream.read(8)

        # solar_radiation
        record_bitstream.read(16)

        # storm_rain
        record_bitstream.read(16)

        # start_date_of_storm
        record_bitstream.read(16)

        # day_rain
        record_bitstream.read(16)

        # month_rain
        record_bitstream.read(16)

        # year_rain
        record_bitstream.read(16)

        # day_et
        record_bitstream.read(16)

        # month_et
        record_bitstream.read(16)

        # year_et
        record_bitstream.read(16)

        # Skip extra soil moistures
        record_bitstream.read(32)
        # Skip extra leaf wetnesses
        record_bitstream.read(32)

        # inside_alarms
        record_bitstream.read(8)

        # rain_alarms
        record_bitstream.read(8)

        # outside_alarms
        record_bitstream.read(16)

        # Skip extra temp/humidity alarms
        record_bitstream.read(64)
        # Skip extra soil leaf alarms
        record_bitstream.read(32)

        # Skip tx battery status
        record_bitstream.read(8)

        console_battery_voltage = (((record_bitstream.read(16).uintle * 300.0) / 512.0) / 100.0)
        forecast_icons = record_bitstream.read(8).uintle
        forecast_rule_number = record_bitstream.read(8).uintle
        sunrise = make_time(record_bitstream.read(16).uintle)
        sunset = make_time(record_bitstream.read(16).uintle)

        return cls(
            bar_trend=bar_trend,
            barometer=barometer,
            inside_temperature=inside_temperature,
            inside_humidity=inside_humidity,
            outside_temperature=outside_temperature,
            outside_humidity=outside_humidity,
            wind_speed=wind_speed,
            wind_direction=wind_direction,
            ten_min_avg_wind_speed=ten_min_avg_wind_speed,
            rain_rate=rain_rate,
            console_battery_voltage=console_battery_voltage,
            forecast_icons=forecast_icons,
            forecast_rule_number=forecast_rule_number,
            sunrise=sunrise,
            sunset=sunset,
            identifier=identifier,
            observation_made_at=observation_made_at)



"""
Highway Conditions
"""
# Standard Library
from collections import defaultdict
import logging
from typing import Dict, List, Optional

# Third Party Code
import requests

SEGMENTS_URL = "https://cotrip.org/speed/getSegments.do"
CONDITIONS_URL = "https://cotrip.org/roadConditions/getRoadConditions.do"

logger = logging.getLogger(__name__)


def get_segments(
    road_name: str,
    start_mile_marker: float,
    end_mile_marker: float,
    direction: Optional[str] = None,
) -> Dict:
    """
    Gets segment conditions
    """
    logging.info(
        "Getting speed segments for %s %s between %s and %s",
        road_name,
        direction or "",
        start_mile_marker,
        end_mile_marker,
    )
    segments_response = requests.get(SEGMENTS_URL)
    segments = defaultdict(list)
    if segments_response.status_code == requests.codes.ok:
        for segment in segments_response.json()["SpeedDetails"]["Segment"]:
            if segment["RoadName"] == road_name:
                segment_start = float(segment["StartMileMarker"])
                segment_end = float(segment["EndMileMarker"])
                if start_mile_marker <= segment_start <= end_mile_marker or (
                    start_mile_marker < segment_end < end_mile_marker
                ):
                    d = {
                        "id": segment["SegmentId"],
                        "road_id": segment["RoadId"],
                        "name": segment["SegmentName"],
                        "direction": segment["Direction"],
                        "start": segment_start,
                        "end": segment_end,
                        "limit": int(segment["SpeedLimit"]),
                        "speed": int(segment["AverageSpeed"]),
                    }
                    if not direction or direction and d["direction"] == direction:
                        segments[d["direction"]].append(d)

    for key in segments.keys():
        if key == "N":
            segments[key].sort(key=lambda k: ["start"])
        else:
            segments[key].sort(key=lambda k: k["start"], reverse=True)
    return dict(segments)


def get_conditions(
    road_name: str, start_mile_marker: float, end_mile_marker: float
) -> List[Dict]:
    """
    Gets the road condtions.
    """
    logging.info(
        "Getting road conditions for %s from %s to %s",
        road_name,
        start_mile_marker,
        end_mile_marker,
    )
    conditions_response = requests.get(CONDITIONS_URL)
    conditions = []
    if conditions_response.status_code == requests.codes.ok:
        for condition in conditions_response.json()["RoadConditionsDetails"][
            "WeatherRoute"
        ]:
            if condition["RoadName"] == road_name:
                # if start_mile_marker <= float(condition["StartMileMarker"]) <= end_mile_marker:
                if start_mile_marker <= float(
                    condition["EndMileMarker"]
                ) and end_mile_marker >= float(condition["StartMileMarker"]):
                    d = {
                        "id": condition["WeatherRouteId"],
                        "road_id": condition["RoadId"],
                        "name": condition["RouteName"],
                        "start": float(condition["StartMileMarker"]),
                        "end": float(condition["EndMileMarker"]),
                        "conditions": condition["RoadCondition"],
                        "is_dangerous": condition["IsHazardousCondition"] == "true"
                        or False,
                    }
                    conditions.append(d)
    return conditions


def get_road_status(
    road_name: str,
    start_mile_marker: float,
    end_mile_marker: float,
    direction: Optional[str],
) -> Dict:
    """
    Retrieves road quality and conditions
    """
    logger.info(
        "Getting road status for %s %s between %s and %s",
        road_name,
        direction or "",
        start_mile_marker,
        end_mile_marker,
    )
    directions = get_segments(
        road_name=road_name,
        start_mile_marker=start_mile_marker,
        end_mile_marker=end_mile_marker,
        direction=direction,
    )
    conditions = get_conditions(
        road_name=road_name,
        start_mile_marker=start_mile_marker,
        end_mile_marker=end_mile_marker,
    )
    for key, items in directions.items():
        for item in items:
            for condition in conditions:
                if (
                    condition["start"] <= item["end"]
                    and condition["end"] >= item["start"]
                ):
                    item["conditions"] = condition["conditions"]
                    item["is_dangerous"] = condition["is_dangerous"]
    return directions



"""
Weather Functions.
"""
# Standard Library
import datetime
import logging
from typing import Dict, List

# Third Party Code
from dateutil.parser import parse
from dateutil.tz import tzlocal
import requests

# Supercell Code
from supercell.models import Forecast, Observation

__version__ = "1.3.1"

WEATHER_URL = (
    "https://api.weather.com/v2/pws/observations/current?stationId="
    "{station_id}&format=json&units=e&apiKey={api_key}"
)
FORECAST_URL = (
    "https://api.weather.com/v3/wx/forecast/daily/5day?geocode="
    "{latitude},{longitude}&format=json&units=e&apiKey={api_key}"
    "&language=en-US"
)
NEAR_URL = (
    "https://api.weather.com/v3/location/near?format=json&geocode="
    "{latitude},{longitude}&product=postal&apiKey={api_key}"
)

logger = logging.getLogger(__name__)


def get_nearby(longitude: float, latitude: float, api_key: str) -> Dict:
    """
    Get a dictionary of nearby locations.
    """
    near_response = requests.get(
        NEAR_URL.format(latitude=latitude, longitude=longitude, api_key=api_key)
    )
    return near_response.json()


def get_forecasts(longitude: float, latitude: float, api_key: str) -> List[Forecast]:
    """
    Gets a list of forecasts for a location.
    """
    forecast_response = requests.get(
        FORECAST_URL.format(latitude=latitude, longitude=longitude, api_key=api_key)
    )
    forecasts = []
    forecast_data = forecast_response.json()
    for i, forecast_for_datetime_str in enumerate(forecast_data["validTimeLocal"]):
        try:
            forecast_for_datetime = parse(forecast_for_datetime_str)
        except ValueError:
            continue

        forecast_for_date = forecast_for_datetime.date()
        forecast_for_utc_offset = forecast_for_datetime.utcoffset()

        if forecast_for_utc_offset:
            forecast_for_utc_offset_seconds = int(
                forecast_for_utc_offset.total_seconds()
            )
        else:
            forecast_for_utc_offset_seconds = 0

        forecast_made = datetime.datetime.now(tzlocal())
        temperature_min = forecast_data["temperatureMin"][i]
        temperature_max = forecast_data["temperatureMax"][i]

        forecasts.append(
            Forecast(
                forecast_made_datetime=forecast_made,
                forecast_for_date=forecast_for_date,
                forecast_for_utc_offset_seconds=forecast_for_utc_offset_seconds,
                temperature_max=temperature_max,
                temperature_min=temperature_min,
                longitude=longitude,
                latitude=latitude,
            )
        )
    return forecasts


def get_station_weather(station_id: str, api_key: str) -> Observation:
    """
    Obtains an observation from a Weather Station.
    """
    weather_response = requests.get(
        WEATHER_URL.format(station_id=station_id, api_key=api_key)
    )
    return Observation(
        latitude=weather_response.json()["observations"][0]["lat"],
        longitude=weather_response.json()["observations"][0]["lon"],
        humidity=weather_response.json()["observations"][0]["humidity"],
        temperature=weather_response.json()["observations"][0]["imperial"]["temp"],
        windchill=weather_response.json()["observations"][0]["imperial"]["windChill"],
        windspeed=weather_response.json()["observations"][0]["imperial"]["windSpeed"],
        pressure=weather_response.json()["observations"][0]["imperial"]["pressure"],
        windgust=weather_response.json()["observations"][0]["imperial"]["windGust"],
        observed_at_str=weather_response.json()["observations"][0]["obsTimeUtc"] + "Z",
    )



"""Fetches GOES images from NOAA."""
# Standard Library
from datetime import date, datetime
import os
from pathlib import Path
import re
from typing import Dict, List, Union
from urllib.parse import urljoin, urlunparse

# Third Party Code
import requests

IMAGE_HOST = "cdn.star.nesdis.noaa.gov"

IMAGE_PATTERN = r"https://{host}(?P<image_path>[" r"a-zA-Z0-9/:\-_.]+)',\s+".format(
    host=re.escape(IMAGE_HOST)
)
IMAGE_COMPPAT = re.compile(IMAGE_PATTERN)

FILE_PATTERN = (
    r"/GOES(?P<sat_num>\d{2})"
    r"/ABI/SECTOR/"
    r"(?P<sector>[a-z]{2,3})/(?P<band>[A-Za-z]+)/"
    r"(?P<t_year>\d{4})"
    r"(?P<t_daynum>\d{3})"
    r"(?P<t_hour>\d{2})"
    r"(?P<t_minute>\d{2})"
    r"_GOES\d{2}-ABI-"
    r"[a-z]{2,3}-[A-Za-z]+-(?P<i_width>\d+)x(?P<i_height>\d+).jpg"
)

IMAGE_COMPOSE = (
    "/GOES{sat_num}/ABI/SECTOR/{sector}/{band}/{t_year}"
    "{t_daynum}{t_hour}{t_minute}_GOES{sat_num}-ABI-{sector}"
    "-{band}-{i_width}x{i_height}.jpg"
)

FILE_COMPPAT = re.compile(FILE_PATTERN)

SATS = {
    "WEST": {"id": "G17", "sectors": {"Pacific Coast": "wus"},},
    "EAST": {
        "id": "G16",
        "sectors": {
            "Northern Rockies": "nr",
            "Upper Mississippi Valley": "umv",
            "Great Lakes": "cgl",
        },
    },
}

GEOCOLOR_BAND = "GEOCOLOR"
SANDWICH_BAND = "Sandwich"

BANDS = {
    GEOCOLOR_BAND,
    SANDWICH_BAND,
}

LENGTHS = {12, 24, 36, 48, 60, 72, 84, 96}


def get_image_data(image_url: str) -> Dict:
    """Produces image data from a URL"""
    mappings = [
        ("t_year", int),
        ("t_daynum", int),
        ("t_hour", str),
        ("t_minute", str),
        ("i_width", int),
        ("i_height", int),
        ("sat_num", int),
        ("sector", str),
        ("band", str),
    ]
    data = FILE_COMPPAT.search(image_url)
    ret = {}
    if data:
        for k, f in mappings:
            match = data.group(k)
            ret[k] = f(match)
    return ret


def does_exist(url: str) -> bool:
    """Determines if a particular file exists on the server."""
    resp = requests.head(url)
    if resp.status_code == requests.codes.ok:
        return True
    return False


def image_url_from_data(image_data: Dict) -> str:
    """Produces a URL from a dictionary of parameters"""
    image_path = IMAGE_COMPOSE.format(
        sat_num=image_data["sat_num"],
        sector=image_data["sector"],
        band=image_data["band"],
        t_year=image_data["t_year"],
        t_daynum=image_data["t_daynum"],
        t_hour=image_data["t_hour"],
        t_minute=image_data["t_minute"],
        i_width=image_data["i_width"],
        i_height=image_data["i_height"],
    )
    return urlunparse(["https", IMAGE_HOST, image_path, None, None, None])


def in_cache(key: str, directory: str) -> Path:
    """Determines if we already have downloaded an image."""
    full_key = Path(directory, key)
    if os.path.exists(full_key):
        return full_key
    raise ValueError("Not in cache.")


def store_in_cache(key: str, data: bytes, directory: str) -> Path:
    """
    Stores some data in a file in a directory.
    """
    full_key = Path(directory, key)
    with open(full_key, "wb") as fh:
        fh.write(data)
    return full_key


def full_day_images(
    sat: str,
    sector: str,
    band: str,
    size: int = 600,
    anchor_datetime: Union[datetime, None] = None,
) -> List:
    """Produces urls for a full day"""
    now = anchor_datetime or datetime.utcnow()
    day_num = int(now.strftime("%j"))
    now_time = int(now.strftime("%H%M"))
    year = now.year
    sat_num = sat[1:]

    def make_url(path: str) -> str:
        """Produces a full URL from a path"""
        return urlunparse(["https", IMAGE_HOST, path, None, None, None])

    index_path = "/GOES{sat_num}/ABI/SECTOR/{sector}/{band}/".format(
        sat_num=sat_num, sector=sector, band=band
    )

    if day_num == 1:  # First day of the new year
        year_two = year - 1
        prev_day_num = (date(year + 1, 1, 1) - date(year, 1, 1)).days
    else:
        year_two = year
        prev_day_num = day_num - 1

    # A pattern which will match all the images corresponding to a specific date.
    pattern_string = (
        r"(?P<file_name>(?P<year>{yearone}|{yeartwo})(?P<day_num>{day_num}|{prev_day_num})(?P<time>["
        r"0-9]{{4}})_GOES{sat_num}-ABI-{sector}-{band}-{size}x{size}.jpg)"
    )
    pattern = re.compile(
        pattern_string.format(
            yearone=year,
            yeartwo=year_two,
            day_num=day_num,
            prev_day_num=prev_day_num,
            sat_num=sat_num,
            sector=sector,
            band=band,
            size=size,
        )
    )
    # Get a list of every image available on the NOAA servers (in HTML)
    index_response = requests.get(make_url(index_path))

    if index_response.status_code == requests.codes.ok:
        # Produce a list of matching image urls for the day and time we care about.
        matches = {
            make_url(urljoin(index_path, i_filename))
            for i_filename, _, i_daynum, i_time in pattern.findall(index_response.text)
            if int(i_daynum) == day_num
            or (int(i_daynum) == prev_day_num and int(i_time) > now_time)
        }
        return sorted(list(matches))
    else:
        return []



"""
Tools for retrieving the current conditions from a Davis Weatherlink IP Logger
"""
# Standard Library
import logging
import socket
import time

# Supercell Code
from supercell.weatherlink.exceptions import BadCRC, NotAcknowledged, UnknownResponseCode
from supercell.weatherlink.models import StationObservation
from supercell.weatherlink.utils import connect, receive_data, request

LOOP_COMMAND = b"LOOP %d\n"
LOOP_RECORD_SIZE_BYTES = 99
LOOP_RECORD_SIZE_BITS = LOOP_RECORD_SIZE_BYTES * 8


logger = logging.getLogger(__name__)


def get_current(host: str, port: int) -> bytes:
    """Gets the current readings on the device."""
    sock = connect(host, port)
    loop_data = b""
    try:
        try:
            request(sock, LOOP_COMMAND % 1)
        except (BadCRC, NotAcknowledged, UnknownResponseCode):
            logger.exception("Could not issue loop command.")
            raise

        while len(loop_data) != LOOP_RECORD_SIZE_BYTES:
            data = receive_data(sock)
            loop_data += data
    except socket.timeout:
        logger.exception("Could not issue loop command")
        raise NotAcknowledged()
    finally:
        sock.close()
    return loop_data


def get_current_condition(host: str, port: int) -> StationObservation:
    """Obtains the current conditions."""
    try:
        current_bytes = get_current(host, port)
    except (BadCRC, NotAcknowledged, UnknownResponseCode):
        # Wait a little and try again.
        time.sleep(0.1)
        try:
            current_bytes = get_current(host, port)
        except (BadCRC, NotAcknowledged, UnknownResponseCode):
            # Wait twice as long and try again.
            time.sleep(1.0)
            try:
                current_bytes = get_current(host, port)
            except (BadCRC, NotAcknowledged, UnknownResponseCode):
                # Ok just give up.
                raise
    return StationObservation.init_with_bytes(current_bytes)



# Standard Library
import datetime
import logging
import socket
from typing import List, Optional

# Third Party Code
from bitstring import BitStream

# Supercell Code
from supercell.weatherlink.exceptions import BadCRC, NotAcknowledged, UnknownResponseCode

socket.setdefaulttimeout(2.0)

SOCKET_BUFFER_SIZE = 16

RESPONSE_CODE_SIZE = 1

NACK_RESPONSE_CODE = 0x21
BAD_CRC_RESPONSE_CODE = 0x18
ACKNOWLEDGED_RESPONSE_CODE = 0x06

CRC16_TABLE = [
    0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
    0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
    0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,
    0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
    0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,
    0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
    0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4,
    0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
    0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823,
    0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
    0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12,
    0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
    0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41,
    0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
    0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70,
    0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
    0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,
    0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
    0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,
    0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
    0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,
    0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
    0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,
    0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
    0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,
    0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,
    0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,
    0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,
    0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,
    0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,
    0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,
    0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0,
]

logger = logging.getLogger(__name__)


def crc16(data: BitStream, crc: int, table: List) -> int:
    """
    Calculate CRC16 using the given table.

    - `data`: The data to calculate the CRC of
    - `crc`: Initial Value
    - `table`: Table for calculating CRC (list of 256 integers)

    Return calculated value of CRC. Should be 0.
    """
    for data_byte in data:
        high_byte = crc >> 8
        low_byte = (crc & 0x00FF) << 8
        table_index = high_byte ^ data_byte
        table_value = table[table_index]
        crc = low_byte ^ table_value
    return crc


def connect(host: str, port: int) -> socket.socket:
    """Connects to a TCP/IP host. """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    logger.info("Connecting to %s:%s", host, port)
    sock.connect((host, port))
    return sock


def request(sock: socket.socket, body: bytes) -> None:
    """Send a request to a socket."""
    sock.sendall(body)
    response_code = BitStream(receive_data(sock, RESPONSE_CODE_SIZE)).int
    if response_code == NACK_RESPONSE_CODE:
        logger.error("Request was not acknowledged.")
        raise NotAcknowledged()
    elif response_code == BAD_CRC_RESPONSE_CODE:
        logger.error("Request contained a bad CRC, retransmit.")
        raise BadCRC()
    elif response_code == ACKNOWLEDGED_RESPONSE_CODE:
        logger.info("Request was acknowledged.")
    else:
        logger.error("Unknown response code %s", response_code)
        raise UnknownResponseCode()


def receive_data(sock: socket.socket, buffer_size: Optional[int] = None) -> bytes:
    """Receives data from a socket."""
    return sock.recv(buffer_size or SOCKET_BUFFER_SIZE)


def make_time(time_stamp: int) -> datetime.time:
    """Converts an integer time to a time object."""
    hour = int(time_stamp / 100)
    minute = time_stamp - (hour * 100)
    return datetime.time(hour=hour, minute=minute)


def make_datetime(date_stamp: BitStream, time_stamp: int) -> Optional[datetime.datetime]:
    """Converts a date stamp and time stamp into a datetime object."""
    year = date_stamp.read(7).uint
    month = date_stamp.read(4).uint
    day = date_stamp.read(5).uint
    timestamp = make_time(time_stamp)
    return datetime.datetime(year=2000 + year, month=month, day=day, hour=timestamp.hour, minute=timestamp.minute)



"""
Various Models.
"""
# Standard Library
import datetime
import random
from typing import Dict, Optional, Union

# Third Party Code
from dateutil.parser import parse
from dateutil.tz import tzutc


class Forecast(object):
    """
    A forecast.
    """

    identifier: int

    def __init__(
        self,
        forecast_for_date: Union[datetime.date, str],
        forecast_for_utc_offset_seconds: int,
        forecast_made_datetime: Union[datetime.datetime, str],
        temperature_min: float,
        temperature_max: float,
        longitude: float,
        latitude: float,
        identifier: Optional[int] = None,
    ) -> None:

        if isinstance(forecast_for_date, datetime.datetime):
            self._forecast_for_date = forecast_for_date.date()
        elif isinstance(forecast_for_date, datetime.date):
            self._forecast_for_date = forecast_for_date
        elif isinstance(forecast_for_date, str):
            self._forecast_for_date = parse(forecast_for_date).date()
        else:
            raise TypeError("Forecast for date must be a string or date.")

        self.forecast_for_utc_offset_seconds = int(forecast_for_utc_offset_seconds)

        if isinstance(forecast_made_datetime, datetime.datetime):
            self.forecast_made_datetime = forecast_made_datetime
        elif isinstance(forecast_made_datetime, str):
            self.forecast_made_datetime = parse(forecast_made_datetime)
        else:
            raise TypeError("Forecast made must be a string or datetime.")

        self.temperature_min = temperature_min and float(temperature_min) or None
        self.temperature_max = temperature_max and float(temperature_max) or None
        self.longitude = longitude
        self.latitude = latitude
        self.identifier = identifier or random.getrandbits(32)

    @property
    def forecast_made_date(self) -> datetime.date:
        return self.forecast_made_datetime.date()

    @property
    def forecast_made_date_float(self) -> float:
        """The date the forecast was made."""
        return datetime.datetime(
            year=self.forecast_made_datetime.year,
            month=self.forecast_made_datetime.month,
            day=self.forecast_made_datetime.day,
            tzinfo=tzutc(),
        ).timestamp()

    @property
    def forecast_for_date(self):
        return datetime.datetime(
            year=self._forecast_for_date.year,
            month=self._forecast_for_date.month,
            day=self._forecast_for_date.day,
            tzinfo=tzutc(),
        ).timestamp()

    @property
    def forecast_made_time(self) -> datetime.time:
        """The time the forecast was made"""
        return self.forecast_made_datetime.timetz()

    @property
    def forecast_made_time_int(self) -> int:
        return (
            self.forecast_made_datetime.hour * 100
        ) + self.forecast_made_datetime.minute

    @property
    def forecast_made_year(self) -> int:
        """The year the forecast was made"""
        return self.forecast_made_datetime.year

    @property
    def forecast_made_month(self) -> int:
        """The month the forecast was made"""
        return self.forecast_made_datetime.month

    @property
    def forecast_made_day(self) -> int:
        """The day the forecast was made"""
        return self.forecast_made_datetime.day

    @property
    def forecast_made_hour(self) -> int:
        """The hour the forecast was made"""
        return self.forecast_made_time.hour

    @property
    def forecast_made_minute(self) -> int:
        """The minute the forecast was made"""
        return self.forecast_made_time.minute

    @property
    def forecast_made_utc_offset_seconds(self) -> int:
        """The UTC offset of the location where the forecast was made."""
        offset = self.forecast_made_datetime.utcoffset()
        if offset:
            return int(offset.total_seconds())
        return 0

    @property
    def forecast_for_year(self) -> int:
        """The year the forecast is for"""
        return self._forecast_for_date.year

    @property
    def forecast_for_month(self) -> int:
        """The month the forecast is for"""
        return self._forecast_for_date.month

    @property
    def forecast_for_day(self) -> int:
        """The day the forecast is for"""
        return self._forecast_for_date.day

    @property
    def forecast_made_str(self) -> str:
        return str(self.forecast_made_datetime)

    @property
    def forecast_made(self) -> float:
        return self.forecast_made_datetime.timestamp()

    @property
    def forecast_for_date_str(self) -> str:
        return self._forecast_for_date.strftime("%Y-%m-%d")

    def to_dict(self) -> Dict:
        """A dictionary representation of the forecast"""
        return {
            "identifier": self.identifier,
            "latitude": self.latitude,
            "longitude": self.longitude,
            "forecast_made_str": self.forecast_made_str,
            "forecast_made": self.forecast_made,
            "forecast_made_date": self.forecast_made_date_float,
            "forecast_made_time": self.forecast_made_time_int,
            "forecast_made_utc_offset_seconds": self.forecast_made_utc_offset_seconds,
            "forecast_for_date_str": self.forecast_for_date_str,
            "forecast_for_date": self.forecast_for_date,
            "forecast_for_utc_offset_seconds": self.forecast_for_utc_offset_seconds,
            "temperature_min": self.temperature_min,
            "temperature_max": self.temperature_max,
        }


class Observation(object):
    """An observation."""

    identifier: int
    humidity: float
    longitude: float
    latitude: float
    windchill: Optional[float]
    pressure: Optional[float]
    windgust: Optional[float]

    def __init__(
        self,
        temperature: float,
        humidity: float,
        longitude: float,
        latitude: float,
        observed_at_str: str,
        windchill: Optional[float] = None,
        windspeed: Optional[float] = None,
        pressure: Optional[float] = None,
        windgust: Optional[float] = None,
        identifier: Optional[int] = None,
    ) -> None:
        self.temperature = float(temperature)
        self.humidity = float(humidity)
        self.windchill = windchill and float(windchill) or None
        self.windspeed = windspeed and float(windspeed) or None
        self.pressure = pressure and float(pressure) or None
        self.windgust = windgust and float(windgust) or None
        self.latitude = float(latitude)
        self.longitude = float(longitude)
        self._observed_at = parse(observed_at_str)
        self.identifier = identifier or random.getrandbits(32)

    @property
    def observed_at_utc_offset_seconds(self) -> int:
        """The UTC offset of the location where the observation happened."""
        offset = self._observed_at.utcoffset()
        if offset:
            return int(offset.total_seconds())
        return 0

    @property
    def observed_at_str(self) -> str:
        return str(self._observed_at)

    @property
    def observed_at(self) -> float:
        return self._observed_at.timestamp()

    @property
    def observed_at_date(self) -> float:
        return datetime.datetime(
            year=self._observed_at.year,
            month=self._observed_at.month,
            day=self._observed_at.day,
            tzinfo=tzutc(),
        ).timestamp()

    @property
    def observed_at_time(self) -> int:
        return (self._observed_at.time().hour * 100) + self._observed_at.time().minute

    def to_dict(self) -> Dict:
        """A dictionary representation of the observation"""
        return {
            "identifier": self.identifier,
            "longitude": self.longitude,
            "latitude": self.latitude,
            "observed_at_str": self.observed_at_str,
            "observed_at": self.observed_at,
            "observed_at_utc_offset_seconds": self.observed_at_utc_offset_seconds,
            "observed_at_date": self.observed_at_date,
            "observed_at_time": self.observed_at_time,
            "temperature": self.temperature,
            "humidity": self.humidity,
        }



class NotAcknowledged(Exception):
    """When the device does not acknowledge a command"""
    pass


class BadCRC(Exception):
    """When the device believes your CRC to be bad."""
    pass


class UnknownResponseCode(Exception):
    """When the device returns an unknown response code."""
    pass



"""
Gets highway signs.
"""
# Standard Library
import functools
import io
import logging
from pathlib import Path
import shutil
from typing import Dict, IO, List, Optional
from urllib.parse import urlunparse

# Third Party Code
import requests

DMS_URL = "https://cotrip.org/device/getDMS.do"
IMAGE_HOST = "i.cotrip.org"

logger = logging.getLogger(__name__)


def get_all_signs() -> List[Dict]:
    """
    Gets all signs.
    """
    response = requests.get(DMS_URL)
    if response.status_code == requests.codes.ok:
        return response.json()["DMSDetails"]["DMS"]
    raise Exception("Couldn't fetch highway data.")


def get_sign_message(dms_id: str, signs: Optional[List[Dict]] = None) -> str:
    """
    Gets a particular sign's image url.
    """
    signs = signs or get_all_signs()
    for s in signs:
        if s["DMSId"] == dms_id:
            return urlunparse(
                ["https", IMAGE_HOST, s["MessageImage"], None, None, None]
            )
    raise ValueError("Could not find that sign.")


def fetch_sign_image(sign_id: str, all_signs: Optional[List[Dict]] = None) -> bytes:
    with requests.get(get_sign_message(sign_id, all_signs), stream=True) as r:
        if r.status_code == requests.codes.ok:
            r.raw.read = functools.partial(r.raw.read, decode_content=True)
            return r.content
    raise ValueError("Could not fetch sign.")


def store_sign_image(data: bytes, local_path: Path) -> None:
    with local_path.open("wb") as f:
        source: IO = io.BytesIO(data)
        dest: IO = f
        shutil.copyfileobj(source, dest)



"""
Highway Cameras
"""
# Standard Library
import functools
import io
import logging
from pathlib import Path
import shutil
from typing import IO, Tuple
from urllib.parse import urlunparse

# Third Party Code
import requests

HOST = "i.cotrip.org"
PATH = "/dimages/camera"
QUERY_PARAM = "imageURL"

logger = logging.getLogger(__name__)


def fetch_cam(cam: str) -> Tuple[str, bytes]:
    """
    Fetches an frame from a camera.
    """
    name, cam_path = cam.split(":")
    logger.info("Downloading Camera %s to %s", cam_path, name)
    image_url = urlunparse(
        ["https", HOST, PATH, None, "%s=%s" % (QUERY_PARAM, cam_path), None]
    )
    with requests.get(image_url, stream=True) as r:
        if r.status_code == requests.codes.ok:
            r.raw.read = functools.partial(r.raw.read, decode_content=True)
            return name, r.content
    raise ValueError("Could not fetch that camera.")


def store_cam(name: str, data: bytes, directory: str) -> Path:
    """
    Stores data to a local file.
    """
    local_path = Path(directory, "highway-cam-%s.jpg" % name)
    with open(local_path, "wb") as f:
        source: IO = io.BytesIO(data)
        dest: IO = f
        shutil.copyfileobj(source, dest)
    return local_path


def download_cam(cam: str, directory: str) -> str:
    """
    Fetches and then stores a camera image.
    """
    name, data = fetch_cam(cam)
    return str(store_cam(name, data, directory).resolve())
